# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mozilla CommonVoice Dataset Builder UnitTest."""

from tensorflow_datasets.audio import commonvoice
import tensorflow_datasets.testing as tfds_test


class CommonVoiceTest(tfds_test.DatasetBuilderTestCase):
  """CommonVoice Tester Class."""
  # Don't test all configs to avoid timeout
  BUILDER_CONFIG_NAMES_TO_TEST = ["en", "cv"]
  DATASET_CLASS = commonvoice.CommonVoice
  SPLITS = {"train": 2, "test": 1, "validation": 3}


if __name__ == "__main__":
  tfds_test.test_main()
