# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for cifar10_1 dataset module."""

from tensorflow_datasets import testing
from tensorflow_datasets.image_classification import cifar10_1
import tensorflow_datasets.public_api as tfds


class Cifar10_1Test(testing.DatasetBuilderTestCase):  # pylint: disable=invalid-name
  DATASET_CLASS = cifar10_1.Cifar10_1
  BUILDER_CONFIG_NAMES_TO_TEST = ["v4"]

  SPLITS = {
      tfds.Split.TEST: 10,
  }

  DL_EXTRACT_RESULT = [
      "cifar10.1_v4_data.npy",
      "cifar10.1_v4_labels.npy",
  ]


if __name__ == "__main__":
  testing.test_main()
