# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Object detection datasets."""

from tensorflow_datasets.object_detection.coco import Coco
from tensorflow_datasets.object_detection.coco_captions import CocoCaptions
from tensorflow_datasets.object_detection.kitti import Kitti
from tensorflow_datasets.object_detection.open_images import OpenImagesV4
from tensorflow_datasets.object_detection.open_images_challenge2019 import OpenImagesChallenge2019Detection
from tensorflow_datasets.object_detection.voc import Voc
from tensorflow_datasets.object_detection.waymo_open_dataset import WaymoOpenDataset
from tensorflow_datasets.object_detection.wider_face import WiderFace

"""Medical Datasets"""
from tensorflow_datasets.object_detection.deeplesion import Deeplesion
