# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: waymo-open-dataset/waymo_open_dataset/dataset.proto
# File was auto-generated:
# pylint: skip-file

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='waymo-open-dataset/waymo_open_dataset/dataset.proto',
  package='waymo.open_dataset',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=b'\n3waymo-open-dataset/waymo_open_dataset/dataset.proto\x12\x12waymo.open_dataset\"\x8f\x06\n\x05Label\x12*\n\x03\x62ox\x18\x01 \x01(\x0b\x32\x1d.waymo.open_dataset.Label.Box\x12\x34\n\x08metadata\x18\x02 \x01(\x0b\x32\".waymo.open_dataset.Label.Metadata\x12,\n\x04type\x18\x03 \x01(\x0e\x32\x1e.waymo.open_dataset.Label.Type\x12\n\n\x02id\x18\x04 \x01(\t\x12M\n\x1a\x64\x65tection_difficulty_level\x18\x05 \x01(\x0e\x32).waymo.open_dataset.Label.DifficultyLevel\x12L\n\x19tracking_difficulty_level\x18\x06 \x01(\x0e\x32).waymo.open_dataset.Label.DifficultyLevel\x12\x1f\n\x17num_lidar_points_in_box\x18\x07 \x01(\x05\x1a\xbf\x01\n\x03\x42ox\x12\x10\n\x08\x63\x65nter_x\x18\x01 \x01(\x01\x12\x10\n\x08\x63\x65nter_y\x18\x02 \x01(\x01\x12\x10\n\x08\x63\x65nter_z\x18\x03 \x01(\x01\x12\x0e\n\x06length\x18\x05 \x01(\x01\x12\r\n\x05width\x18\x04 \x01(\x01\x12\x0e\n\x06height\x18\x06 \x01(\x01\x12\x0f\n\x07heading\x18\x07 \x01(\x01\"B\n\x04Type\x12\x10\n\x0cTYPE_UNKNOWN\x10\x00\x12\x0b\n\x07TYPE_3D\x10\x01\x12\x0b\n\x07TYPE_2D\x10\x02\x12\x0e\n\nTYPE_AA_2D\x10\x03\x1aN\n\x08Metadata\x12\x0f\n\x07speed_x\x18\x01 \x01(\x01\x12\x0f\n\x07speed_y\x18\x02 \x01(\x01\x12\x0f\n\x07\x61\x63\x63\x65l_x\x18\x03 \x01(\x01\x12\x0f\n\x07\x61\x63\x63\x65l_y\x18\x04 \x01(\x01\"`\n\x04Type\x12\x10\n\x0cTYPE_UNKNOWN\x10\x00\x12\x10\n\x0cTYPE_VEHICLE\x10\x01\x12\x13\n\x0fTYPE_PEDESTRIAN\x10\x02\x12\r\n\tTYPE_SIGN\x10\x03\x12\x10\n\x0cTYPE_CYCLIST\x10\x04\"8\n\x0f\x44ifficultyLevel\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07LEVEL_1\x10\x01\x12\x0b\n\x07LEVEL_2\x10\x02\"2\n\x0ePolygon2dProto\x12\t\n\x01x\x18\x01 \x03(\x01\x12\t\n\x01y\x18\x02 \x03(\x01\x12\n\n\x02id\x18\x03 \x01(\t\"\x1b\n\x0bMatrixShape\x12\x0c\n\x04\x64ims\x18\x01 \x03(\x05\"O\n\x0bMatrixFloat\x12\x10\n\x04\x64\x61ta\x18\x01 \x03(\x02\x42\x02\x10\x01\x12.\n\x05shape\x18\x02 \x01(\x0b\x32\x1f.waymo.open_dataset.MatrixShape\"O\n\x0bMatrixInt32\x12\x10\n\x04\x64\x61ta\x18\x01 \x03(\x05\x42\x02\x10\x01\x12.\n\x05shape\x18\x02 \x01(\x0b\x32\x1f.waymo.open_dataset.MatrixShape\"l\n\nCameraName\"^\n\x04Name\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05\x46RONT\x10\x01\x12\x0e\n\nFRONT_LEFT\x10\x02\x12\x0f\n\x0b\x46RONT_RIGHT\x10\x03\x12\r\n\tSIDE_LEFT\x10\x04\x12\x0e\n\nSIDE_RIGHT\x10\x05\"]\n\tLaserName\"P\n\x04Name\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x07\n\x03TOP\x10\x01\x12\t\n\x05\x46RONT\x10\x02\x12\r\n\tSIDE_LEFT\x10\x03\x12\x0e\n\nSIDE_RIGHT\x10\x04\x12\x08\n\x04REAR\x10\x05\"\x1e\n\tTransform\x12\x11\n\ttransform\x18\x01 \x03(\x01\"X\n\x08Velocity\x12\x0b\n\x03v_x\x18\x01 \x01(\x02\x12\x0b\n\x03v_y\x18\x02 \x01(\x02\x12\x0b\n\x03v_z\x18\x03 \x01(\x02\x12\x0b\n\x03w_x\x18\x04 \x01(\x01\x12\x0b\n\x03w_y\x18\x05 \x01(\x01\x12\x0b\n\x03w_z\x18\x06 \x01(\x01\"\xa3\x03\n\x11\x43\x61meraCalibration\x12\x31\n\x04name\x18\x01 \x01(\x0e\x32#.waymo.open_dataset.CameraName.Name\x12\x11\n\tintrinsic\x18\x02 \x03(\x01\x12\x30\n\textrinsic\x18\x03 \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\x12\r\n\x05width\x18\x04 \x01(\x05\x12\x0e\n\x06height\x18\x05 \x01(\x05\x12g\n\x19rolling_shutter_direction\x18\x06 \x01(\x0e\x32\x44.waymo.open_dataset.CameraCalibration.RollingShutterReadOutDirection\"\x8d\x01\n\x1eRollingShutterReadOutDirection\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x11\n\rTOP_TO_BOTTOM\x10\x01\x12\x11\n\rLEFT_TO_RIGHT\x10\x02\x12\x11\n\rBOTTOM_TO_TOP\x10\x03\x12\x11\n\rRIGHT_TO_LEFT\x10\x04\x12\x12\n\x0eGLOBAL_SHUTTER\x10\x05\"\xcd\x01\n\x10LaserCalibration\x12\x30\n\x04name\x18\x01 \x01(\x0e\x32\".waymo.open_dataset.LaserName.Name\x12\x19\n\x11\x62\x65\x61m_inclinations\x18\x02 \x03(\x01\x12\x1c\n\x14\x62\x65\x61m_inclination_min\x18\x03 \x01(\x01\x12\x1c\n\x14\x62\x65\x61m_inclination_max\x18\x04 \x01(\x01\x12\x30\n\textrinsic\x18\x05 \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\"\xf6\x03\n\x07\x43ontext\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x42\n\x13\x63\x61mera_calibrations\x18\x02 \x03(\x0b\x32%.waymo.open_dataset.CameraCalibration\x12@\n\x12laser_calibrations\x18\x03 \x03(\x0b\x32$.waymo.open_dataset.LaserCalibration\x12\x30\n\x05stats\x18\x04 \x01(\x0b\x32!.waymo.open_dataset.Context.Stats\x1a\xa4\x02\n\x05Stats\x12J\n\x13laser_object_counts\x18\x01 \x03(\x0b\x32-.waymo.open_dataset.Context.Stats.ObjectCount\x12K\n\x14\x63\x61mera_object_counts\x18\x05 \x03(\x0b\x32-.waymo.open_dataset.Context.Stats.ObjectCount\x12\x13\n\x0btime_of_day\x18\x02 \x01(\t\x12\x10\n\x08location\x18\x03 \x01(\t\x12\x0f\n\x07weather\x18\x04 \x01(\t\x1aJ\n\x0bObjectCount\x12,\n\x04type\x18\x01 \x01(\x0e\x32\x1e.waymo.open_dataset.Label.Type\x12\r\n\x05\x63ount\x18\x02 \x01(\x05\"\xb1\x01\n\nRangeImage\x12\x1e\n\x16range_image_compressed\x18\x02 \x01(\x0c\x12$\n\x1c\x63\x61mera_projection_compressed\x18\x03 \x01(\x0c\x12#\n\x1brange_image_pose_compressed\x18\x04 \x01(\x0c\x12\x38\n\x0brange_image\x18\x01 \x01(\x0b\x32\x1f.waymo.open_dataset.MatrixFloatB\x02\x18\x01\"\x94\x02\n\x0b\x43\x61meraImage\x12\x31\n\x04name\x18\x01 \x01(\x0e\x32#.waymo.open_dataset.CameraName.Name\x12\r\n\x05image\x18\x02 \x01(\x0c\x12+\n\x04pose\x18\x03 \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\x12.\n\x08velocity\x18\x04 \x01(\x0b\x32\x1c.waymo.open_dataset.Velocity\x12\x16\n\x0epose_timestamp\x18\x05 \x01(\x01\x12\x0f\n\x07shutter\x18\x06 \x01(\x01\x12\x1b\n\x13\x63\x61mera_trigger_time\x18\x07 \x01(\x01\x12 \n\x18\x63\x61mera_readout_done_time\x18\x08 \x01(\x01\"l\n\x0c\x43\x61meraLabels\x12\x31\n\x04name\x18\x01 \x01(\x0e\x32#.waymo.open_dataset.CameraName.Name\x12)\n\x06labels\x18\x02 \x03(\x0b\x32\x19.waymo.open_dataset.Label\"\xa1\x01\n\x05Laser\x12\x30\n\x04name\x18\x01 \x01(\x0e\x32\".waymo.open_dataset.LaserName.Name\x12\x32\n\nri_return1\x18\x02 \x01(\x0b\x32\x1e.waymo.open_dataset.RangeImage\x12\x32\n\nri_return2\x18\x03 \x01(\x0b\x32\x1e.waymo.open_dataset.RangeImage\"\xc0\x03\n\x05\x46rame\x12,\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x1b.waymo.open_dataset.Context\x12\x18\n\x10timestamp_micros\x18\x02 \x01(\x03\x12+\n\x04pose\x18\x03 \x01(\x0b\x32\x1d.waymo.open_dataset.Transform\x12/\n\x06images\x18\x04 \x03(\x0b\x32\x1f.waymo.open_dataset.CameraImage\x12)\n\x06lasers\x18\x05 \x03(\x0b\x32\x19.waymo.open_dataset.Laser\x12/\n\x0claser_labels\x18\x06 \x03(\x0b\x32\x19.waymo.open_dataset.Label\x12@\n\x16projected_lidar_labels\x18\t \x03(\x0b\x32 .waymo.open_dataset.CameraLabels\x12\x37\n\rcamera_labels\x18\x08 \x03(\x0b\x32 .waymo.open_dataset.CameraLabels\x12:\n\x0eno_label_zones\x18\x07 \x03(\x0b\x32\".waymo.open_dataset.Polygon2dProto'
)



_LABEL_BOX_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='waymo.open_dataset.Label.Box.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPE_3D', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPE_2D', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPE_AA_2D', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=557,
  serialized_end=623,
)
_sym_db.RegisterEnumDescriptor(_LABEL_BOX_TYPE)

_LABEL_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='waymo.open_dataset.Label.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPE_VEHICLE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPE_PEDESTRIAN', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPE_SIGN', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPE_CYCLIST', index=4, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=705,
  serialized_end=801,
)
_sym_db.RegisterEnumDescriptor(_LABEL_TYPE)

_LABEL_DIFFICULTYLEVEL = _descriptor.EnumDescriptor(
  name='DifficultyLevel',
  full_name='waymo.open_dataset.Label.DifficultyLevel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LEVEL_1', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LEVEL_2', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=803,
  serialized_end=859,
)
_sym_db.RegisterEnumDescriptor(_LABEL_DIFFICULTYLEVEL)

_CAMERANAME_NAME = _descriptor.EnumDescriptor(
  name='Name',
  full_name='waymo.open_dataset.CameraName.Name',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FRONT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FRONT_LEFT', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FRONT_RIGHT', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIDE_LEFT', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIDE_RIGHT', index=5, number=5,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1118,
  serialized_end=1212,
)
_sym_db.RegisterEnumDescriptor(_CAMERANAME_NAME)

_LASERNAME_NAME = _descriptor.EnumDescriptor(
  name='Name',
  full_name='waymo.open_dataset.LaserName.Name',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOP', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FRONT', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIDE_LEFT', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIDE_RIGHT', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REAR', index=5, number=5,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1227,
  serialized_end=1307,
)
_sym_db.RegisterEnumDescriptor(_LASERNAME_NAME)

_CAMERACALIBRATION_ROLLINGSHUTTERREADOUTDIRECTION = _descriptor.EnumDescriptor(
  name='RollingShutterReadOutDirection',
  full_name='waymo.open_dataset.CameraCalibration.RollingShutterReadOutDirection',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOP_TO_BOTTOM', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LEFT_TO_RIGHT', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BOTTOM_TO_TOP', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RIGHT_TO_LEFT', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GLOBAL_SHUTTER', index=5, number=5,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1710,
  serialized_end=1851,
)
_sym_db.RegisterEnumDescriptor(_CAMERACALIBRATION_ROLLINGSHUTTERREADOUTDIRECTION)


_LABEL_BOX = _descriptor.Descriptor(
  name='Box',
  full_name='waymo.open_dataset.Label.Box',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='center_x', full_name='waymo.open_dataset.Label.Box.center_x', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='center_y', full_name='waymo.open_dataset.Label.Box.center_y', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='center_z', full_name='waymo.open_dataset.Label.Box.center_z', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='length', full_name='waymo.open_dataset.Label.Box.length', index=3,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='waymo.open_dataset.Label.Box.width', index=4,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='waymo.open_dataset.Label.Box.height', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='heading', full_name='waymo.open_dataset.Label.Box.heading', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LABEL_BOX_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=432,
  serialized_end=623,
)

_LABEL_METADATA = _descriptor.Descriptor(
  name='Metadata',
  full_name='waymo.open_dataset.Label.Metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='speed_x', full_name='waymo.open_dataset.Label.Metadata.speed_x', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='speed_y', full_name='waymo.open_dataset.Label.Metadata.speed_y', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='accel_x', full_name='waymo.open_dataset.Label.Metadata.accel_x', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='accel_y', full_name='waymo.open_dataset.Label.Metadata.accel_y', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=625,
  serialized_end=703,
)

_LABEL = _descriptor.Descriptor(
  name='Label',
  full_name='waymo.open_dataset.Label',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='box', full_name='waymo.open_dataset.Label.box', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='waymo.open_dataset.Label.metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='waymo.open_dataset.Label.type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='waymo.open_dataset.Label.id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='detection_difficulty_level', full_name='waymo.open_dataset.Label.detection_difficulty_level', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tracking_difficulty_level', full_name='waymo.open_dataset.Label.tracking_difficulty_level', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_lidar_points_in_box', full_name='waymo.open_dataset.Label.num_lidar_points_in_box', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_LABEL_BOX, _LABEL_METADATA, ],
  enum_types=[
    _LABEL_TYPE,
    _LABEL_DIFFICULTYLEVEL,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=76,
  serialized_end=859,
)


_POLYGON2DPROTO = _descriptor.Descriptor(
  name='Polygon2dProto',
  full_name='waymo.open_dataset.Polygon2dProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='waymo.open_dataset.Polygon2dProto.x', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='y', full_name='waymo.open_dataset.Polygon2dProto.y', index=1,
      number=2, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='waymo.open_dataset.Polygon2dProto.id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=861,
  serialized_end=911,
)


_MATRIXSHAPE = _descriptor.Descriptor(
  name='MatrixShape',
  full_name='waymo.open_dataset.MatrixShape',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dims', full_name='waymo.open_dataset.MatrixShape.dims', index=0,
      number=1, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=913,
  serialized_end=940,
)


_MATRIXFLOAT = _descriptor.Descriptor(
  name='MatrixFloat',
  full_name='waymo.open_dataset.MatrixFloat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='waymo.open_dataset.MatrixFloat.data', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shape', full_name='waymo.open_dataset.MatrixFloat.shape', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=942,
  serialized_end=1021,
)


_MATRIXINT32 = _descriptor.Descriptor(
  name='MatrixInt32',
  full_name='waymo.open_dataset.MatrixInt32',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='waymo.open_dataset.MatrixInt32.data', index=0,
      number=1, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shape', full_name='waymo.open_dataset.MatrixInt32.shape', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1023,
  serialized_end=1102,
)


_CAMERANAME = _descriptor.Descriptor(
  name='CameraName',
  full_name='waymo.open_dataset.CameraName',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CAMERANAME_NAME,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1104,
  serialized_end=1212,
)


_LASERNAME = _descriptor.Descriptor(
  name='LaserName',
  full_name='waymo.open_dataset.LaserName',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LASERNAME_NAME,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1214,
  serialized_end=1307,
)


_TRANSFORM = _descriptor.Descriptor(
  name='Transform',
  full_name='waymo.open_dataset.Transform',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transform', full_name='waymo.open_dataset.Transform.transform', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1309,
  serialized_end=1339,
)


_VELOCITY = _descriptor.Descriptor(
  name='Velocity',
  full_name='waymo.open_dataset.Velocity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='v_x', full_name='waymo.open_dataset.Velocity.v_x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v_y', full_name='waymo.open_dataset.Velocity.v_y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='v_z', full_name='waymo.open_dataset.Velocity.v_z', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='w_x', full_name='waymo.open_dataset.Velocity.w_x', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='w_y', full_name='waymo.open_dataset.Velocity.w_y', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='w_z', full_name='waymo.open_dataset.Velocity.w_z', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1341,
  serialized_end=1429,
)


_CAMERACALIBRATION = _descriptor.Descriptor(
  name='CameraCalibration',
  full_name='waymo.open_dataset.CameraCalibration',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='waymo.open_dataset.CameraCalibration.name', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='intrinsic', full_name='waymo.open_dataset.CameraCalibration.intrinsic', index=1,
      number=2, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extrinsic', full_name='waymo.open_dataset.CameraCalibration.extrinsic', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='waymo.open_dataset.CameraCalibration.width', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='waymo.open_dataset.CameraCalibration.height', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rolling_shutter_direction', full_name='waymo.open_dataset.CameraCalibration.rolling_shutter_direction', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CAMERACALIBRATION_ROLLINGSHUTTERREADOUTDIRECTION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1432,
  serialized_end=1851,
)


_LASERCALIBRATION = _descriptor.Descriptor(
  name='LaserCalibration',
  full_name='waymo.open_dataset.LaserCalibration',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='waymo.open_dataset.LaserCalibration.name', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beam_inclinations', full_name='waymo.open_dataset.LaserCalibration.beam_inclinations', index=1,
      number=2, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beam_inclination_min', full_name='waymo.open_dataset.LaserCalibration.beam_inclination_min', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beam_inclination_max', full_name='waymo.open_dataset.LaserCalibration.beam_inclination_max', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extrinsic', full_name='waymo.open_dataset.LaserCalibration.extrinsic', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1854,
  serialized_end=2059,
)


_CONTEXT_STATS_OBJECTCOUNT = _descriptor.Descriptor(
  name='ObjectCount',
  full_name='waymo.open_dataset.Context.Stats.ObjectCount',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='waymo.open_dataset.Context.Stats.ObjectCount.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='count', full_name='waymo.open_dataset.Context.Stats.ObjectCount.count', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2490,
  serialized_end=2564,
)

_CONTEXT_STATS = _descriptor.Descriptor(
  name='Stats',
  full_name='waymo.open_dataset.Context.Stats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='laser_object_counts', full_name='waymo.open_dataset.Context.Stats.laser_object_counts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='camera_object_counts', full_name='waymo.open_dataset.Context.Stats.camera_object_counts', index=1,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='time_of_day', full_name='waymo.open_dataset.Context.Stats.time_of_day', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='location', full_name='waymo.open_dataset.Context.Stats.location', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weather', full_name='waymo.open_dataset.Context.Stats.weather', index=4,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CONTEXT_STATS_OBJECTCOUNT, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2272,
  serialized_end=2564,
)

_CONTEXT = _descriptor.Descriptor(
  name='Context',
  full_name='waymo.open_dataset.Context',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='waymo.open_dataset.Context.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='camera_calibrations', full_name='waymo.open_dataset.Context.camera_calibrations', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='laser_calibrations', full_name='waymo.open_dataset.Context.laser_calibrations', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stats', full_name='waymo.open_dataset.Context.stats', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CONTEXT_STATS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2062,
  serialized_end=2564,
)


_RANGEIMAGE = _descriptor.Descriptor(
  name='RangeImage',
  full_name='waymo.open_dataset.RangeImage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='range_image_compressed', full_name='waymo.open_dataset.RangeImage.range_image_compressed', index=0,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='camera_projection_compressed', full_name='waymo.open_dataset.RangeImage.camera_projection_compressed', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='range_image_pose_compressed', full_name='waymo.open_dataset.RangeImage.range_image_pose_compressed', index=2,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='range_image', full_name='waymo.open_dataset.RangeImage.range_image', index=3,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2567,
  serialized_end=2744,
)


_CAMERAIMAGE = _descriptor.Descriptor(
  name='CameraImage',
  full_name='waymo.open_dataset.CameraImage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='waymo.open_dataset.CameraImage.name', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='image', full_name='waymo.open_dataset.CameraImage.image', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pose', full_name='waymo.open_dataset.CameraImage.pose', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='velocity', full_name='waymo.open_dataset.CameraImage.velocity', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pose_timestamp', full_name='waymo.open_dataset.CameraImage.pose_timestamp', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shutter', full_name='waymo.open_dataset.CameraImage.shutter', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='camera_trigger_time', full_name='waymo.open_dataset.CameraImage.camera_trigger_time', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='camera_readout_done_time', full_name='waymo.open_dataset.CameraImage.camera_readout_done_time', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2747,
  serialized_end=3023,
)


_CAMERALABELS = _descriptor.Descriptor(
  name='CameraLabels',
  full_name='waymo.open_dataset.CameraLabels',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='waymo.open_dataset.CameraLabels.name', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='labels', full_name='waymo.open_dataset.CameraLabels.labels', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3025,
  serialized_end=3133,
)


_LASER = _descriptor.Descriptor(
  name='Laser',
  full_name='waymo.open_dataset.Laser',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='waymo.open_dataset.Laser.name', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ri_return1', full_name='waymo.open_dataset.Laser.ri_return1', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ri_return2', full_name='waymo.open_dataset.Laser.ri_return2', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3136,
  serialized_end=3297,
)


_FRAME = _descriptor.Descriptor(
  name='Frame',
  full_name='waymo.open_dataset.Frame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='context', full_name='waymo.open_dataset.Frame.context', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp_micros', full_name='waymo.open_dataset.Frame.timestamp_micros', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pose', full_name='waymo.open_dataset.Frame.pose', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='images', full_name='waymo.open_dataset.Frame.images', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lasers', full_name='waymo.open_dataset.Frame.lasers', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='laser_labels', full_name='waymo.open_dataset.Frame.laser_labels', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='projected_lidar_labels', full_name='waymo.open_dataset.Frame.projected_lidar_labels', index=6,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='camera_labels', full_name='waymo.open_dataset.Frame.camera_labels', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='no_label_zones', full_name='waymo.open_dataset.Frame.no_label_zones', index=8,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3300,
  serialized_end=3748,
)

_LABEL_BOX.containing_type = _LABEL
_LABEL_BOX_TYPE.containing_type = _LABEL_BOX
_LABEL_METADATA.containing_type = _LABEL
_LABEL.fields_by_name['box'].message_type = _LABEL_BOX
_LABEL.fields_by_name['metadata'].message_type = _LABEL_METADATA
_LABEL.fields_by_name['type'].enum_type = _LABEL_TYPE
_LABEL.fields_by_name['detection_difficulty_level'].enum_type = _LABEL_DIFFICULTYLEVEL
_LABEL.fields_by_name['tracking_difficulty_level'].enum_type = _LABEL_DIFFICULTYLEVEL
_LABEL_TYPE.containing_type = _LABEL
_LABEL_DIFFICULTYLEVEL.containing_type = _LABEL
_MATRIXFLOAT.fields_by_name['shape'].message_type = _MATRIXSHAPE
_MATRIXINT32.fields_by_name['shape'].message_type = _MATRIXSHAPE
_CAMERANAME_NAME.containing_type = _CAMERANAME
_LASERNAME_NAME.containing_type = _LASERNAME
_CAMERACALIBRATION.fields_by_name['name'].enum_type = _CAMERANAME_NAME
_CAMERACALIBRATION.fields_by_name['extrinsic'].message_type = _TRANSFORM
_CAMERACALIBRATION.fields_by_name['rolling_shutter_direction'].enum_type = _CAMERACALIBRATION_ROLLINGSHUTTERREADOUTDIRECTION
_CAMERACALIBRATION_ROLLINGSHUTTERREADOUTDIRECTION.containing_type = _CAMERACALIBRATION
_LASERCALIBRATION.fields_by_name['name'].enum_type = _LASERNAME_NAME
_LASERCALIBRATION.fields_by_name['extrinsic'].message_type = _TRANSFORM
_CONTEXT_STATS_OBJECTCOUNT.fields_by_name['type'].enum_type = _LABEL_TYPE
_CONTEXT_STATS_OBJECTCOUNT.containing_type = _CONTEXT_STATS
_CONTEXT_STATS.fields_by_name['laser_object_counts'].message_type = _CONTEXT_STATS_OBJECTCOUNT
_CONTEXT_STATS.fields_by_name['camera_object_counts'].message_type = _CONTEXT_STATS_OBJECTCOUNT
_CONTEXT_STATS.containing_type = _CONTEXT
_CONTEXT.fields_by_name['camera_calibrations'].message_type = _CAMERACALIBRATION
_CONTEXT.fields_by_name['laser_calibrations'].message_type = _LASERCALIBRATION
_CONTEXT.fields_by_name['stats'].message_type = _CONTEXT_STATS
_RANGEIMAGE.fields_by_name['range_image'].message_type = _MATRIXFLOAT
_CAMERAIMAGE.fields_by_name['name'].enum_type = _CAMERANAME_NAME
_CAMERAIMAGE.fields_by_name['pose'].message_type = _TRANSFORM
_CAMERAIMAGE.fields_by_name['velocity'].message_type = _VELOCITY
_CAMERALABELS.fields_by_name['name'].enum_type = _CAMERANAME_NAME
_CAMERALABELS.fields_by_name['labels'].message_type = _LABEL
_LASER.fields_by_name['name'].enum_type = _LASERNAME_NAME
_LASER.fields_by_name['ri_return1'].message_type = _RANGEIMAGE
_LASER.fields_by_name['ri_return2'].message_type = _RANGEIMAGE
_FRAME.fields_by_name['context'].message_type = _CONTEXT
_FRAME.fields_by_name['pose'].message_type = _TRANSFORM
_FRAME.fields_by_name['images'].message_type = _CAMERAIMAGE
_FRAME.fields_by_name['lasers'].message_type = _LASER
_FRAME.fields_by_name['laser_labels'].message_type = _LABEL
_FRAME.fields_by_name['projected_lidar_labels'].message_type = _CAMERALABELS
_FRAME.fields_by_name['camera_labels'].message_type = _CAMERALABELS
_FRAME.fields_by_name['no_label_zones'].message_type = _POLYGON2DPROTO
DESCRIPTOR.message_types_by_name['Label'] = _LABEL
DESCRIPTOR.message_types_by_name['Polygon2dProto'] = _POLYGON2DPROTO
DESCRIPTOR.message_types_by_name['MatrixShape'] = _MATRIXSHAPE
DESCRIPTOR.message_types_by_name['MatrixFloat'] = _MATRIXFLOAT
DESCRIPTOR.message_types_by_name['MatrixInt32'] = _MATRIXINT32
DESCRIPTOR.message_types_by_name['CameraName'] = _CAMERANAME
DESCRIPTOR.message_types_by_name['LaserName'] = _LASERNAME
DESCRIPTOR.message_types_by_name['Transform'] = _TRANSFORM
DESCRIPTOR.message_types_by_name['Velocity'] = _VELOCITY
DESCRIPTOR.message_types_by_name['CameraCalibration'] = _CAMERACALIBRATION
DESCRIPTOR.message_types_by_name['LaserCalibration'] = _LASERCALIBRATION
DESCRIPTOR.message_types_by_name['Context'] = _CONTEXT
DESCRIPTOR.message_types_by_name['RangeImage'] = _RANGEIMAGE
DESCRIPTOR.message_types_by_name['CameraImage'] = _CAMERAIMAGE
DESCRIPTOR.message_types_by_name['CameraLabels'] = _CAMERALABELS
DESCRIPTOR.message_types_by_name['Laser'] = _LASER
DESCRIPTOR.message_types_by_name['Frame'] = _FRAME
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Label = _reflection.GeneratedProtocolMessageType('Label', (_message.Message,), {

  'Box' : _reflection.GeneratedProtocolMessageType('Box', (_message.Message,), {
    'DESCRIPTOR' : _LABEL_BOX,
    '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
    # @@protoc_insertion_point(class_scope:waymo.open_dataset.Label.Box)
    })
  ,

  'Metadata' : _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), {
    'DESCRIPTOR' : _LABEL_METADATA,
    '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
    # @@protoc_insertion_point(class_scope:waymo.open_dataset.Label.Metadata)
    })
  ,
  'DESCRIPTOR' : _LABEL,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.Label)
  })
_sym_db.RegisterMessage(Label)
_sym_db.RegisterMessage(Label.Box)
_sym_db.RegisterMessage(Label.Metadata)

Polygon2dProto = _reflection.GeneratedProtocolMessageType('Polygon2dProto', (_message.Message,), {
  'DESCRIPTOR' : _POLYGON2DPROTO,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.Polygon2dProto)
  })
_sym_db.RegisterMessage(Polygon2dProto)

MatrixShape = _reflection.GeneratedProtocolMessageType('MatrixShape', (_message.Message,), {
  'DESCRIPTOR' : _MATRIXSHAPE,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.MatrixShape)
  })
_sym_db.RegisterMessage(MatrixShape)

MatrixFloat = _reflection.GeneratedProtocolMessageType('MatrixFloat', (_message.Message,), {
  'DESCRIPTOR' : _MATRIXFLOAT,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.MatrixFloat)
  })
_sym_db.RegisterMessage(MatrixFloat)

MatrixInt32 = _reflection.GeneratedProtocolMessageType('MatrixInt32', (_message.Message,), {
  'DESCRIPTOR' : _MATRIXINT32,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.MatrixInt32)
  })
_sym_db.RegisterMessage(MatrixInt32)

CameraName = _reflection.GeneratedProtocolMessageType('CameraName', (_message.Message,), {
  'DESCRIPTOR' : _CAMERANAME,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.CameraName)
  })
_sym_db.RegisterMessage(CameraName)

LaserName = _reflection.GeneratedProtocolMessageType('LaserName', (_message.Message,), {
  'DESCRIPTOR' : _LASERNAME,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.LaserName)
  })
_sym_db.RegisterMessage(LaserName)

Transform = _reflection.GeneratedProtocolMessageType('Transform', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFORM,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.Transform)
  })
_sym_db.RegisterMessage(Transform)

Velocity = _reflection.GeneratedProtocolMessageType('Velocity', (_message.Message,), {
  'DESCRIPTOR' : _VELOCITY,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.Velocity)
  })
_sym_db.RegisterMessage(Velocity)

CameraCalibration = _reflection.GeneratedProtocolMessageType('CameraCalibration', (_message.Message,), {
  'DESCRIPTOR' : _CAMERACALIBRATION,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.CameraCalibration)
  })
_sym_db.RegisterMessage(CameraCalibration)

LaserCalibration = _reflection.GeneratedProtocolMessageType('LaserCalibration', (_message.Message,), {
  'DESCRIPTOR' : _LASERCALIBRATION,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.LaserCalibration)
  })
_sym_db.RegisterMessage(LaserCalibration)

Context = _reflection.GeneratedProtocolMessageType('Context', (_message.Message,), {

  'Stats' : _reflection.GeneratedProtocolMessageType('Stats', (_message.Message,), {

    'ObjectCount' : _reflection.GeneratedProtocolMessageType('ObjectCount', (_message.Message,), {
      'DESCRIPTOR' : _CONTEXT_STATS_OBJECTCOUNT,
      '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
      # @@protoc_insertion_point(class_scope:waymo.open_dataset.Context.Stats.ObjectCount)
      })
    ,
    'DESCRIPTOR' : _CONTEXT_STATS,
    '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
    # @@protoc_insertion_point(class_scope:waymo.open_dataset.Context.Stats)
    })
  ,
  'DESCRIPTOR' : _CONTEXT,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.Context)
  })
_sym_db.RegisterMessage(Context)
_sym_db.RegisterMessage(Context.Stats)
_sym_db.RegisterMessage(Context.Stats.ObjectCount)

RangeImage = _reflection.GeneratedProtocolMessageType('RangeImage', (_message.Message,), {
  'DESCRIPTOR' : _RANGEIMAGE,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.RangeImage)
  })
_sym_db.RegisterMessage(RangeImage)

CameraImage = _reflection.GeneratedProtocolMessageType('CameraImage', (_message.Message,), {
  'DESCRIPTOR' : _CAMERAIMAGE,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.CameraImage)
  })
_sym_db.RegisterMessage(CameraImage)

CameraLabels = _reflection.GeneratedProtocolMessageType('CameraLabels', (_message.Message,), {
  'DESCRIPTOR' : _CAMERALABELS,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.CameraLabels)
  })
_sym_db.RegisterMessage(CameraLabels)

Laser = _reflection.GeneratedProtocolMessageType('Laser', (_message.Message,), {
  'DESCRIPTOR' : _LASER,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.Laser)
  })
_sym_db.RegisterMessage(Laser)

Frame = _reflection.GeneratedProtocolMessageType('Frame', (_message.Message,), {
  'DESCRIPTOR' : _FRAME,
  '__module__' : 'waymo_open_dataset.waymo_open_dataset.dataset_pb2'
  # @@protoc_insertion_point(class_scope:waymo.open_dataset.Frame)
  })
_sym_db.RegisterMessage(Frame)


_MATRIXFLOAT.fields_by_name['data']._options = None
_MATRIXINT32.fields_by_name['data']._options = None
_RANGEIMAGE.fields_by_name['range_image']._options = None
# @@protoc_insertion_point(module_scope)
